from typing import Dict
import numpy as np  # type: ignore
from inewave.config import NUM_CENARIOS, MESES


class Earmfpm00:
    """
    Classe responsável por lidar com o armazenamento dos dados das
    saídas do NWListOP referentes às energias armazenadas finais,
    por submercado e em % da energia armazenável máxima:
    earmfpm00x.out.
    """
    def __init__(self,
                 mes_pmo: int,
                 ano_pmo: int,
                 versao_newave: str,
                 submercado: str,
                 energias_armazenadas: Dict[int,
                                            np.ndarray]):
        self.mes_pmo = mes_pmo
        self.ano_pmo = ano_pmo
        self.versao_newave = versao_newave
        self.submercado = submercado
        self.energias_armazenadas = energias_armazenadas

    @property
    def energias_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Retorna as energias armazenadas para cada ano e em cada cenário, para
        todos os meses, organizadas primeiramente por ano. O acesso é feito
        com [ano] e retorna um np.ndarray com os valores de EARM para todos os
        cenários e meses, naquele ano.
        """
        return self.energias_armazenadas

    @property
    def energias_por_ano_e_mes(self) -> Dict[int,
                                             Dict[int, np.ndarray]]:
        """
        Retorna as energias armazenadas para cada ano e mês em cada cenário,
        para todos os meses, organizadas primeiramente por ano. O acesso é
        feito com [ano][mes] e retorna um np.ndarray com os valores de EARM
        para todos os cenários, naquele ano e mês.
        """
        energias: Dict[int,
                       Dict[int, np.ndarray]] = {}
        anos_estudo = list(self.energias_armazenadas.keys())
        n_meses = len(MESES)
        # Cria e inicializa os objetos a serem retornados
        for a in anos_estudo:
            energias[a] = {m: None
                           for m in range(1, n_meses + 1)}
        # Preenche com os valores
        for a, tabela in self.energias_armazenadas.items():
            for m in range(1, n_meses + 1):
                col = m - 1
                energias[a][m] = tabela[:, col]
        return energias

    @property
    def energias_por_ano_e_cenario(self) -> Dict[int,
                                                 Dict[int, np.ndarray]]:
        """
        Retorna as energias armazenadas para cada ano e cenário, para todos os
        meses, organizadas primeiramente por ano. O acesso é feito
        com [ano][cenario] e retorna um np.ndarray com os valores de EARM para
        todos os meses, naquele ano e cenário.
        """
        energias: Dict[int,
                       Dict[int, np.ndarray]] = {}
        anos_estudo = list(self.energias_armazenadas.keys())
        # Cria e inicializa os objetos a serem retornados
        for a in anos_estudo:
            energias[a] = {c: None
                           for c in range(NUM_CENARIOS)}
        # Preenche com os valores
        for a, tabela in self.energias_armazenadas.items():
            for c in range(NUM_CENARIOS):
                energias[a][c] = tabela[c, :]
        return energias



from typing import List, Dict
import numpy as np  # type: ignore
from inewave.config import MESES


class Patamar:
    """
    Classe responsável por lidar com o armazenamento dos dados de
    entrada do NEWAVE referentes aos patamares de carga por submercado,
    localizados no arquivo patamar.dat.
    """
    def __init__(self,
                 num_patamares: int,
                 anos_estudo: List[int],
                 patamares: np.ndarray):
        self.num_patamares = num_patamares
        self.anos_estudo = anos_estudo
        self.patamares = patamares

    @property
    def patamares_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Fornece os valores contidos na tabela de patamares, organizados
        por ano. O acesso é feito com [ano] e o valor fornecido
        é uma array 2-D do numpy com os valores dos patamares para todo
        os meses de um ano, semelhante ao formato do arquivo patamar.dat.
        """
        patamares_ano: Dict[int, np.ndarray] = {}
        # Preenche com os valores
        for i, a in enumerate(self.anos_estudo):
            li = self.num_patamares * i
            lf = li + self.num_patamares
            patamares_ano[a] = self.patamares[li:lf, :]
        return patamares_ano

    @property
    def patamares_por_ano_e_mes(self) -> Dict[int, Dict[int, np.ndarray]]:
        """
        Fornece os valores contidos na tabela de patamares, organizados
        por ano e mês. O acesso é feito com [ano][mes] e o valor fornecido
        é uma array do numpy com os valores para os patamares 1, 2, ... .
        """
        patamares_ano: Dict[int, Dict[int, np.ndarray]] = {}
        n_meses = len(MESES)
        # Cria e inicializa os objetos
        for a in self.anos_estudo:
            patamares_ano[a] = {m: np.zeros(self.num_patamares,)
                                for m in range(1, n_meses + 1)}
        # Preenche com os valores
        for i, a in enumerate(self.anos_estudo):
            for m in range(1, n_meses + 1):
                li = self.num_patamares * i
                lf = li + self.num_patamares
                col = m - 1
                patamares_ano[a][m] = self.patamares[li:lf, col]
        return patamares_ano



# Imports do próprio módulo
from inewave._utils.leitura import Leitura
from inewave.config import NUM_ANOS_ESTUDO, NUM_VARIAVEIS_CUSTO_PMO
from inewave.config import MESES, SUBMERCADOS
from .modelos.pmo import DadosGeraisPMO
from .modelos.pmo import EnergiasAfluentesPMO
from .modelos.pmo import RiscoDeficitENSPMO
from .modelos.pmo import CustoOperacaoPMO
from .modelos.pmo import PMO
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, List, Tuple


class LeituraPMO(Leitura):
    """
    Classe para realizar a leitura do arquivo pmo.dat
    existente em um diretório de saídas do NEWAVE.
    """
    str_dados_pmo = " DATA : "
    str_inicio_risco = " ANO  RISCO   EENS  RISCO"
    str_inicio_custo_series = "                 CUSTO DE OPERACAO DAS"
    str_inicio_valor_esperado = "                 VALOR ESPERADO PARA PERI"
    str_inicio_custo_referenciado = "                     CUSTO OPERACAO R"
    str_fim_pmo = "DETECTADO NO CALCULO DA SIMULACAO FINAL"

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # PMO default, depois é substituído
        self.pmo = PMO(0,
                       0,
                       "",
                       DadosGeraisPMO(),
                       EnergiasAfluentesPMO(),
                       EnergiasAfluentesPMO(),
                       EnergiasAfluentesPMO(),
                       {},
                       RiscoDeficitENSPMO([], np.array([])),
                       CustoOperacaoPMO(np.array([])),
                       CustoOperacaoPMO(np.array([])),
                       CustoOperacaoPMO(np.array([])))

    def le_arquivo(self):
        try:
            caminho = os.path.join(self.diretorio, "pmo.dat")
            with open(caminho, "r") as arq:
                self.pmo = self._le_pmo(arq)
                return self.pmo
        except Exception:
            print_exc()
            return self.pmo

    def _le_pmo(self, arq: IO) -> PMO:
        """
        Faz a leitura do arquivo pmo.dat.
        """
        achou_dados_pmo = False
        leu_dados_pmo = False
        achou_risco_ens = False
        achou_custo_series = False
        achou_valor_esperado = False
        achou_custo_refer = False
        linha = ""
        # Variáveis para armazenar os componentes do PMO, que será
        # construído quando acabar a leitura
        ano_pmo = 0
        mes_pmo = 0
        versao_newave = ""
        risco_ens = RiscoDeficitENSPMO([], np.array([]))
        custo_series = CustoOperacaoPMO(np.array([]))
        valor_esp = CustoOperacaoPMO(np.array([]))
        custo_ref = CustoOperacaoPMO(np.array([]))
        while True:
            # Decide se lê uma linha nova ou usa a última lida
            linha = self._le_linha_com_backup(arq)
            if len(linha) == 0:
                break
            # Condição para finalizar a leitura de um arquivo
            if self._fim_arquivo(linha):
                self.pmo = PMO(ano_pmo,
                               mes_pmo,
                               versao_newave,
                               DadosGeraisPMO(),
                               EnergiasAfluentesPMO(),
                               EnergiasAfluentesPMO(),
                               EnergiasAfluentesPMO(),
                               {},
                               risco_ens,
                               custo_series,
                               valor_esp,
                               custo_ref)
                break
            # Condição para iniciar uma leitura de dados
            if not achou_dados_pmo and not leu_dados_pmo:
                achou = LeituraPMO.str_dados_pmo in linha
                achou_dados_pmo = achou
            if not achou_risco_ens:
                achou = LeituraPMO.str_inicio_risco in linha
                achou_risco_ens = achou
            if not achou_custo_series:
                achou = LeituraPMO.str_inicio_custo_series in linha
                achou_custo_series = achou
            if not achou_valor_esperado:
                achou = LeituraPMO.str_inicio_valor_esperado in linha
                achou_valor_esperado = achou
            if not achou_custo_refer:
                achou = LeituraPMO.str_inicio_custo_referenciado in linha
                achou_custo_refer = achou
            # Quando achar, le cada parte adequadamente
            if achou_dados_pmo:
                ano_pmo, mes_pmo, versao_newave = self._le_dados_pmo(linha)
                achou_dados_pmo = False
                leu_dados_pmo = True
            if achou_risco_ens:
                risco_ens = self._le_risco_ens(arq)
                achou_risco_ens = False
            if achou_custo_series:
                custo_series = self._le_tabela_custo(arq)
                achou_custo_series = False
            if achou_valor_esperado:
                valor_esp = self._le_tabela_custo(arq)
                achou_valor_esperado = False
            if achou_custo_refer:
                custo_ref = self._le_tabela_custo(arq)
                achou_custo_refer = False

        return self.pmo

    def _le_dados_pmo(self, linha: str) -> Tuple[int, int, str]:
        """
        Lê a linha com dados do PMO em questão e retorna dados
        sobre o mês e ano de estudo e a versão do NEWAVE usada.
        """
        # Procura o mês do PMO
        mes_pmo = 0
        for i, m in enumerate(MESES):
            if m in linha:
                mes_pmo = i + 1
                break
        # Quebra a linha no nome do mês
        ano_pmo = int(linha.split(MESES[mes_pmo-1])[1][3:7])
        # Encontra a versao do NW
        versao_newave = linha.split("Versao")[1].strip()
        return ano_pmo, mes_pmo, versao_newave

    def _le_risco_ens(self, arq: IO) -> RiscoDeficitENSPMO:
        """
        Lê as linhas que formam a tabela de RISCO e ENS por ano
        de estudo e subsistema.
        """
        # Salta duas linhas após a string de início ser identificada
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Inicializa as variáveis de interesse
        anos_estudo: List[int] = []
        linhas_tabela = NUM_ANOS_ESTUDO
        colunas_tabela = 2 * len(SUBMERCADOS)
        tabela = np.zeros((linhas_tabela, colunas_tabela))
        campos_colunas = [5, 7] * len(SUBMERCADOS)
        # Lê a tabela
        for i in range(linhas_tabela):
            linha = self._le_linha_com_backup(arq)
            anos_estudo.append(int(linha[1:5]))
            ci = 7
            for j in range(colunas_tabela):
                cf = ci + campos_colunas[j]
                tabela[i, j] = float(linha[ci:cf])
                ci = cf + 1
        # Constroi o objeto e retorna
        return RiscoDeficitENSPMO(anos_estudo, tabela)

    def _le_tabela_custo(self, arq: IO) -> np.ndarray:
        """
        Lê as linhas que formam uma das tabelas de composição
        do valor total de operação do pmo.dat.
        """
        inicio_tabela = "PARCELA           V.ESPERADO"
        iniciou = False
        # Procura o cabeçalho da tabela
        linha = ""
        while not iniciou:
            linha = self._le_linha_com_backup(arq)
            iniciou = inicio_tabela in linha
        # Pula uma linha
        self._le_linha_com_backup(arq)
        # Cria a variável para armazenar a tabela
        tabela = np.zeros((NUM_VARIAVEIS_CUSTO_PMO, 3))
        # Lê cada uma das variáveis
        abs_i = 32
        abs_f = 45
        perc_i = 46
        dp_i = 47
        dp_f = 59
        perc_i = 60
        perc_f = 67
        for i in range(NUM_VARIAVEIS_CUSTO_PMO):
            linha = self._le_linha_com_backup(arq)
            tabela[i, 0] = float(linha[abs_i:abs_f])
            tabela[i, 1] = float(linha[dp_i:dp_f])
            tabela[i, 2] = float(linha[perc_i:perc_f])
        return CustoOperacaoPMO(tabela)

    def _fim_arquivo(self, linha: str) -> bool:
        return LeituraPMO.str_fim_pmo in linha



SUBMERCADOS = ["SUDESTE", "SUL", "NORDESTE", "NORTE"]
NUM_CENARIOS = 2000
NUM_ANOS_ESTUDO = 5
NUM_PATAMARES = 3
NUM_VARIAVEIS_MEDIAS = 52
NUM_VARIAVEIS_CUSTO_PMO = 16
MESES = ["JANEIRO",
         "FEVEREIRO",
         "MARCO",
         "ABRIL",
         "MAIO",
         "JUNHO",
         "JULHO",
         "AGOSTO",
         "SETEMBRO",
         "OUTUBRO",
         "NOVEMBRO",
         "DEZEMBRO"]



# Imports do próprio módulo
from inewave.nwlistop.modelos.mediassin import MediasSIN
from inewave._utils.leitura import Leitura
from inewave.config import NUM_ANOS_ESTUDO, MESES
from inewave.config import NUM_VARIAVEIS_MEDIAS
# Imports de módulos externos
import os
import csv
import numpy as np  # type: ignore
from traceback import print_exc


class LeituraMediasSIN(Leitura):
    """
    Classe para realizar a leitura do arquivo MEDIAS-SIN.CSV
    existente em um diretório de saídas do NEWAVE.
    """
    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # MediasSIN default, depois é substituído
        self.medias = MediasSIN(0, np.array([]))

    def le_arquivo(self) -> MediasSIN:
        try:
            caminho = os.path.join(self.diretorio, "MEDIAS-SIN.CSV")
            n_meses = len(MESES)
            linhas_medias = NUM_VARIAVEIS_MEDIAS + 3
            colunas_medias = NUM_ANOS_ESTUDO * n_meses
            mes_pmo = 0
            with open(caminho, newline='') as arq:
                leitor = csv.reader(arq, delimiter=",", quotechar='|')
                tabela = np.zeros((linhas_medias, colunas_medias))
                # Descobre o mês do PMO pela terceira coluna
                # da primeira linha. Depois lê linha a linha.
                primeira_linha = True
                for i, linha in enumerate(leitor):
                    if primeira_linha:
                        mes_pmo = int(linha[2])
                        primeira_linha = False
                        continue
                    # Ignora as linhas após os dados do SIN
                    if i > linhas_medias:
                        break
                    linha_num = np.array([float(n) for n in linha[2:-1]])
                    tabela[i - 1, mes_pmo-1:] = linha_num
                self.medias = MediasSIN(mes_pmo, tabela)
                return self.medias
        except Exception:
            print_exc()
            return MediasSIN(0, {})

    def _fim_arquivo(self, linha: str) -> bool:
        return False



from typing import Dict, List
import numpy as np  # type: ignore


class DadosGeraisPMO:
    """
    Classe responsável por armazenar as informações de dados
    gerais contidas no arquivo pmo.dat.
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        # Todas as informações echo do dger.dat.
        pass


class EnergiasAfluentesPMO:
    """
    Classe responsável por armazenar as informações de energias
    afluentes anteriores ao estudo contidas no arquivo pmo.dat.
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        pass


class DemandaLiquidaEnergiaPMO:
    """
    Classe responsável por armazenar as informações de demandas
    líquidas de energia contidas no arquivo pmo.dat.
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        pass


class RiscoDeficitENSPMO:
    """
    Classe responsável por armazenar as informações risco de déficit
    e valores esperados de energia não supridacontidas no arquivo pmo.dat.
    """
    def __init__(self,
                 anos_estudo: List[int],
                 tabela: np.ndarray):
        # São impressos no pmo.dat os subsistemas em
        # uma ordem pré-determinada
        self.subsistemas = ["SUDESTE",
                            "SUL",
                            "NORDESTE",
                            "NORTE"]
        self.anos_estudo = anos_estudo
        self.tabela = tabela

    @property
    def riscos_por_subsistema_e_ano(self) -> Dict[str,
                                                  Dict[int,
                                                       float]]:
        """
        Representação dos riscos de déficit agrupados por
        subsistema e ano de estudo. O acesso é feito com
        [sub][ano] e o valor fornecido é em percentual, de 0 a 100.
        """
        riscos: Dict[str, Dict[int, float]] = {}
        for j, sub in enumerate(self.subsistemas):
            if sub not in riscos:
                riscos[sub] = {}
            for i, ano in enumerate(self.anos_estudo):
                riscos[sub][ano] = self.tabela[i, 2*j]
        return riscos

    @property
    def ens_por_subsistema_e_ano(self) -> Dict[str,
                                               Dict[int,
                                                    float]]:
        """
        Representação das energias não supridas agrupadas por
        subsistema e ano de estudo. O acesso é feito com
        [sub][ano] e o valor fornecido é em MWmes.
        """
        energias: Dict[str, Dict[int, float]] = {}
        for j, sub in enumerate(self.subsistemas):
            if sub not in energias:
                energias[sub] = {}
            for i, ano in enumerate(self.anos_estudo):
                energias[sub][ano] = self.tabela[i, 2*j + 1]
        return energias


class CustoOperacaoPMO:
    """
    Classe responsável por armazenar as informações do relatório
    de custo de operação, disponível no arquivo pmo.dat.
    """
    def __init__(self,
                 custos: np.ndarray):
        self.custos = custos
        # TODO - fazer as @property de cada uma individualmente

        # self.geracao_termica = geracao_termica
        # self.deficit = deficit
        # self.vertimento = vertimento
        # self.excesso_energia = excesso_energia
        # self.violacao_car = violacao_car
        # self.violacao_sar = violacao_sar
        # self.violacao_outro_usos = violacao_outro_usos
        # self.violacao_evmin = violacao_evmin
        # self.violacao_vzmin = violacao_vzmin
        # self.intercambio = intercambio
        # self.violacao_intercambio_min = violacao_intercambio_min
        # self.vertimento_fio_nao_turbin = vertimento_fio_nao_turbin
        # self.violacao_ghmin = violacao_ghmin
        # self.violacao_ghmin_usina = violacao_ghmin_usina
        # self.violacao_retirada = violacao_retirada
        # self.violacao_emissao_gee = violacao_emissao_gee

    @property
    def geracao_termica(self) -> float:
        """
        Parcela do custo de operação devido à geração térmica,
        em MMR$.
        """
        return float(self.custos[0, 0])

    @property
    def deficit(self) -> float:
        """
        Parcela do custo de operação devido ao déficit, em MMR$.
        """
        return float(self.custos[1, 0])

    @property
    def custos_seguranca(self) -> float:
        """
        Parcela do custo de operação devido a violações de segurança,
        CAR e SAR, em MMR$.
        """
        return np.sum(self.custos[4:6, 0])

    @property
    def custos_hidricos(self) -> float:
        """
        Parcela do custo de operação devido a violações de restrições
        hídricas, EVmin, VZmin, GHmin, GHmin Usina e outros usos
        da água, em MMR$.
        """
        return (np.sum(self.custos[6:9, 0]) +
                np.sum(self.custos[12:14, 0]))

    @property
    def outros_custos(self) -> float:
        """
        Parcela do custo de operação devido a violações e penalidades
        de intercâmbio, intercâmbio mínimo, vertimento, vertimento
        fio d'água não turbinável, excesso de energia, emissão de GEE
        e retirada, em MMR$.
        """
        return (np.sum(self.custos[2:4, 0]) +
                np.sum(self.custos[9:12, 0]) +
                np.sum(self.custos[14:, 0]))

    @property
    def custo_total(self) -> float:
        """
        Custo total de operação, em MMR$.
        """
        return np.sum(self.custos[:, 0])


class PMO:
    """
    Classe que armazena as informações estruturadas armazenadas no
    arquivo de saída do NEWAVE pmo.dat.
    """
    def __init__(self,
                 ano_pmo: int,
                 mes_pmo: int,
                 versao_newave: str,
                 dados_gerais: DadosGeraisPMO,
                 energias_passadas_politica: EnergiasAfluentesPMO,
                 energias_passadas_primeira_conf: EnergiasAfluentesPMO,
                 energias_passadas_canal_fuga: EnergiasAfluentesPMO,
                 demanda_liquida_energia: Dict[str,
                                               DemandaLiquidaEnergiaPMO],
                 risco_ens: RiscoDeficitENSPMO,
                 custo_series_simuladas: CustoOperacaoPMO,
                 valor_esperado_periodo: CustoOperacaoPMO,
                 custo_referenciado: CustoOperacaoPMO):
        self.ano_pmo = ano_pmo
        self.mes_pmo = mes_pmo
        self.versao_newave = versao_newave
        self.dados_gerais = dados_gerais
        self.energias_passadas_politica = energias_passadas_politica
        self.energias_passadas_primeira_conf = energias_passadas_primeira_conf
        self.energias_passadas_canal_fuga = energias_passadas_canal_fuga
        self.demanda_liquida_energia = demanda_liquida_energia
        self.risco_ens = risco_ens
        self.custo_series_simuladas = custo_series_simuladas
        self.valor_esperado_periodo = valor_esperado_periodo
        self.custo_referenciado = custo_referenciado
